@@ -1,6 +1,6 @@
 ##############################################################################
 #
-# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
+# Copyright (c) 2001, 2002 Zope Corporation and Contributors.
 # All Rights Reserved.
 #
 # This software is subject to the provisions of the Zope Public License,
@@ -14,39 +14,25 @@
 import os, unittest
 import transaction
 import ZODB.FileStorage
+import ZODB.tests.testblob
 import ZODB.tests.util
+import zope.testing.setupstack
 from ZODB import POSException
 from ZODB import DB
 
 from ZODB.tests import StorageTestBase, BasicStorage, TransactionalUndoStorage
-from ZODB.tests import VersionStorage, TransactionalUndoVersionStorage
 from ZODB.tests import PackableStorage, Synchronization, ConflictResolution
 from ZODB.tests import HistoryStorage, IteratorStorage, Corruption
 from ZODB.tests import RevisionStorage, PersistentStorage, MTStorage
 from ZODB.tests import ReadOnlyStorage, RecoveryStorage
 from ZODB.tests.StorageTestBase import MinPO, zodb_pickle
 
-class BaseFileStorageTests(StorageTestBase.StorageTestBase):
-
-    def open(self, **kwargs):
-        self._storage = ZODB.FileStorage.FileStorage('FileStorageTests.fs',
-                                                     **kwargs)
-
-    def setUp(self):
-        self.open(create=1)
-
-    def tearDown(self):
-        self._storage.close()
-        self._storage.cleanup()
-
 class FileStorageTests(
-    BaseFileStorageTests,
+    StorageTestBase.StorageTestBase,
     BasicStorage.BasicStorage,
     TransactionalUndoStorage.TransactionalUndoStorage,
     RevisionStorage.RevisionStorage,
-    VersionStorage.VersionStorage,
-    TransactionalUndoVersionStorage.TransactionalUndoVersionStorage,
-    PackableStorage.PackableStorage,
+    PackableStorage.PackableStorageWithOptionalGC,
     PackableStorage.PackableUndoStorage,
     Synchronization.SynchronizedStorage,
     ConflictResolution.ConflictResolvingStorage,
@@ -59,6 +45,14 @@ class FileStorageTests(
     ReadOnlyStorage.ReadOnlyStorage
     ):
 
+    def open(self, **kwargs):
+        self._storage = ZODB.FileStorage.FileStorage('FileStorageTests.fs',
+                                                     **kwargs)
+
+    def setUp(self):
+        StorageTestBase.StorageTestBase.setUp(self)
+        self.open(create=1)
+
     def checkLongMetadata(self):
         s = "X" * 75000
         try:
@@ -182,45 +176,6 @@ class FileStorageTests(
         self.open()
         self.assertEqual(self._storage._saved, 1)
 
-    def check_index_oid_ignored(self):
-        # Prior to ZODB 3.2.6, the 'oid' value stored in the .index file
-        # was believed.  But there were cases where adding larger oids
-        # didn't update the FileStorage ._oid attribute -- the restore()
-        # method in particular didn't update it, and that's about the only
-        # method copyTransactionsFrom() uses.  A database copy created that
-        # way then stored an 'oid' of z64 in the .index file.  This created
-        # torturous problems, as when that file was opened, "new" oids got
-        # generated starting over from 0 again.
-        # Now the cached 'oid' value is ignored:  verify that this is so.
-        import cPickle as pickle
-        from ZODB.utils import z64
-
-        # Create some data.
-        db = DB(self._storage)
-        conn = db.open()
-        conn.root()['xyz'] = 1
-        transaction.commit()
-        true_max_oid = self._storage._oid
-
-        # Save away the index, and poke in a bad 'oid' value by hand.
-        db.close()
-        f = open('FileStorageTests.fs.index', 'r+b')
-        p = pickle.Unpickler(f)
-        data = p.load()
-        saved_oid = data['oid']
-        self.assertEqual(true_max_oid, saved_oid)
-        data['oid'] = z64
-        f.seek(0)
-        f.truncate()
-        p = pickle.Pickler(f, 1)
-        p.dump(data)
-        f.close()
-
-        # Verify that we get the correct oid again when we reopen, despite
-        # that we stored nonsense in the .index file's 'oid'.
-        self.open()
-        self.assertEqual(self._storage._oid, true_max_oid)
-
     def checkStoreBumpsOid(self):
         # If .store() is handed an oid bigger than the storage knows
         # about already, it's crucial that the storage bump its notion
@@ -264,6 +219,7 @@ class FileStorageTests(
 
         from ZODB.utils import U64, p64
         from ZODB.FileStorage.format import CorruptedError
+        from ZODB.serialize import referencesf
 
         db = DB(self._storage)
         conn = db.open()
@@ -296,7 +252,7 @@ class FileStorageTests(
         # Try to pack.  This used to yield
         #     NameError: global name 's' is not defined
         try:
-            self._storage.pack(time.time(), None)
+            self._storage.pack(time.time(), referencesf)
         except CorruptedError, detail:
             self.assert_("redundant transaction length does not match "
                          "initial transaction length" in str(detail))
@@ -330,11 +286,13 @@ class FileStorageTests(
             else:
                 self.assertNotEqual(next_oid, None)
 
-    def checkOmitVersionOnLoadAndHistory(self):
-        db = ZODB.DB(self._storage)
-        self.assertEqual(self._storage.load('\0'*8),
-                         self._storage.load('\0'*8, ''))
-        self._storage.history('\0'*8)
+class FileStorageTestsWithBlobsEnabled(FileStorageTests):
+
+    def open(self, **kwargs):
+        if 'blob_dir' not in kwargs:
+            kwargs = kwargs.copy()
+            kwargs['blob_dir'] = 'blobs'
+        return FileStorageTests.open(self, **kwargs)
 
 class FileStorageRecoveryTest(
     StorageTestBase.StorageTestBase,
@@ -342,36 +300,108 @@ class FileStorageRecoveryTest(
     ):
 
     def setUp(self):
+        StorageTestBase.StorageTestBase.setUp(self)
         self._storage = ZODB.FileStorage.FileStorage("Source.fs", create=True)
         self._dst = ZODB.FileStorage.FileStorage("Dest.fs", create=True)
 
     def tearDown(self):
-        self._storage.close()
         self._dst.close()
-        self._storage.cleanup()
-        self._dst.cleanup()
+        StorageTestBase.StorageTestBase.tearDown(self)
 
     def new_dest(self):
         return ZODB.FileStorage.FileStorage('Dest.fs')
 
-class SlowFileStorageTest(BaseFileStorageTests):
 
-    level = 2
+class FileStorageNoRestore(ZODB.FileStorage.FileStorage):
 
-    def check10Kstores(self):
-        # The _get_cached_serial() method has a special case
-        # every 8000 calls.  Make sure it gets minimal coverage.
-        oids = [[self._storage.new_oid(), None] for i in range(100)]
-        for i in range(100):
+    @property
+    def restore(self):
+        raise Exception
+
+
+class FileStorageNoRestoreRecoveryTest(FileStorageRecoveryTest):
+    # This test actually verifies a code path of
+    # BaseStorage.copyTransactionsFrom. For simplicity of implementation, we
+    # use a FileStorage deprived of its restore method.
+
+    def setUp(self):
+        StorageTestBase.StorageTestBase.setUp(self)
+        self._storage = FileStorageNoRestore("Source.fs", create=True)
+        self._dst = FileStorageNoRestore("Dest.fs", create=True)
+
+    def new_dest(self):
+        return FileStorageNoRestore('Dest.fs')
+
+    def checkRestoreAcrossPack(self):
+        # Skip this check as it calls restore directly.
+        pass
+
+
+class AnalyzeDotPyTest(StorageTestBase.StorageTestBase):
+
+    def setUp(self):
+        StorageTestBase.StorageTestBase.setUp(self)
+        self._storage = ZODB.FileStorage.FileStorage("Source.fs", create=True)
+
+    def checkanalyze(self):
+        import new, sys, pickle
+        from BTrees.OOBTree import OOBTree
+        from ZODB.scripts import analyze
+        
+        # Set up a module to act as a broken import
+        module_name = 'brokenmodule'
+        module = new.module(module_name)
+        sys.modules[module_name] = module
+        
+        class Broken(MinPO):
+            __module__ = module_name
+        module.Broken = Broken
+        
+        oids = [[self._storage.new_oid(), None] for i in range(3)]
+        for i in range(2):
             t = transaction.Transaction()
             self._storage.tpc_begin(t)
-            for j in range(100):
-                o = MinPO(j)
-                oid, revid = oids[j]
-                serial = self._storage.store(oid, revid, zodb_pickle(o), "", t)
-                oids[j][1] = serial
+            
+            # sometimes data is in this format
+            j = 0
+            oid, revid = oids[j]
+            serial = self._storage.store(oid, revid, pickle.dumps(OOBTree, 1), "", t) 
+            oids[j][1] = serial
+            
+            # and it could be from a broken module
+            j = 1
+            oid, revid = oids[j]
+            serial = self._storage.store(oid, revid, pickle.dumps(Broken, 1), "", t) 
+            oids[j][1] = serial
+            
+            # but mostly it looks like this
+            j = 2
+            o = MinPO(j)
+            oid, revid = oids[j]
+            serial = self._storage.store(oid, revid, zodb_pickle(o), "", t)
+            oids[j][1] = serial
+
             self._storage.tpc_vote(t)
             self._storage.tpc_finish(t)
+            
+        # now break the import of the Broken class
+        del sys.modules[module_name]
+        
+        # from ZODB.scripts.analyze.analyze
+        fsi = self._storage.iterator()
+        rep = analyze.Report()
+        for txn in fsi:
+            analyze.analyze_trans(rep, txn)
+        
+        # from ZODB.scripts.analyze.report
+        typemap = rep.TYPEMAP.keys()
+        typemap.sort()
+        cumpct = 0.0
+        for t in typemap:
+            pct = rep.TYPESIZE[t] * 100.0 / rep.DBYTES
+            cumpct += pct
+       
+        self.assertAlmostEqual(cumpct, 100.0, 0, "Failed to analyze some records")
 
 # Raise an exception if the tids in FileStorage fs aren't
 # strictly increasing.
@@ -379,7 +409,7 @@ def checkIncreasingTids(fs):
     lasttid = '\0' * 8
     for txn in fs.iterator():
         if lasttid >= txn.tid:
-            raise ValueError("tids out of order %r >= %r" % (lasttid, tid))
+            raise ValueError("tids out of order %r >= %r" % (lasttid, txn.tid))
         lasttid = txn.tid
 
 # Return a TimeStamp object 'minutes' minutes in the future.
@@ -480,10 +510,10 @@ track of the transactions along the way:
     >>> fs = ZODB.FileStorage.FileStorage('t.fs', create=True)
     >>> db = DB(fs)
     >>> conn = db.open()
-    >>> from persistent.dict import PersistentDict
+    >>> from persistent.mapping import PersistentMapping
     >>> last = []
     >>> for i in range(100):
-    ...     conn.root()[i] = PersistentDict()
+    ...     conn.root()[i] = PersistentMapping()
     ...     transaction.commit()
     ...     last.append(fs.lastTransaction())
 
@@ -494,7 +524,7 @@ Now, we can call lastInvalidations on it:
     True
 
     >>> from ZODB.utils import u64
-    >>> [[int(u64(oid)) for (oid, version) in oids]
+    >>> [[int(u64(oid)) for oid in oids]
     ...  for (i, oids) in invalidations]
     ... # doctest: +NORMALIZE_WHITESPACE
     [[0, 91], [0, 92], [0, 93], [0, 94], [0, 95],
@@ -508,16 +538,17 @@ there are:
 
 Of course, calling lastInvalidations on an empty storage refturns no data:
 
-    >>> fs.close()
+    >>> db.close()
     >>> fs = ZODB.FileStorage.FileStorage('t.fs', create=True)
     >>> list(fs.lastInvalidations(10))
     []
 
+    >>> fs.close()
     """
 
 def deal_with_finish_failures():
     r"""
-    
+
     It's really bad to get errors in FileStorage's _finish method, as
     that can cause the file storage to be in an inconsistent
     state. The data file will be fine, but the internal data
@@ -546,7 +577,7 @@ def deal_with_finish_failures():
     ...
     TypeError: <lambda>() takes no arguments (1 given)
 
-    
+
     >>> print handler
     ZODB.FileStorage CRITICAL
       Failure in _finish. Closing.
@@ -569,15 +600,62 @@ def deal_with_finish_failures():
     >>> db.close()
     """
 
+def pack_with_open_blob_files():
+    """
+    Make sure packing works while there are open blob files.
+
+    >>> fs = ZODB.FileStorage.FileStorage('data.fs', blob_dir='blobs')
+    >>> db = ZODB.DB(fs)
+    >>> tm1 = transaction.TransactionManager()
+    >>> conn1 = db.open(tm1)
+    >>> import ZODB.blob
+    >>> conn1.root()[1] = ZODB.blob.Blob()
+    >>> conn1.add(conn1.root()[1])
+    >>> conn1.root()[1].open('w').write('some data')
+    >>> tm1.commit()
+
+    >>> tm2 = transaction.TransactionManager()
+    >>> conn2 = db.open(tm2)
+    >>> f = conn1.root()[1].open()
+    >>> conn1.root()[2] = ZODB.blob.Blob()
+    >>> conn1.add(conn1.root()[2])
+    >>> conn1.root()[2].open('w').write('some more data')
+
+    >>> db.pack()
+    >>> f.read()
+    'some data'
+    >>> f.close()
+
+    >>> tm1.commit()
+    >>> conn2.sync()
+    >>> conn2.root()[2].open().read()
+    'some more data'
+
+    >>> db.close()
+    """
+
 def test_suite():
     from zope.testing import doctest
 
     suite = unittest.TestSuite()
     for klass in [FileStorageTests, Corruption.FileStorageCorruptTests,
-                  FileStorageRecoveryTest, SlowFileStorageTest]:
+                  FileStorageRecoveryTest, FileStorageNoRestoreRecoveryTest,
+                  FileStorageTestsWithBlobsEnabled, AnalyzeDotPyTest,
+                  ]:
         suite.addTest(unittest.makeSuite(klass, "check"))
-    suite.addTest(doctest.DocTestSuite(setUp=ZODB.tests.util.setUp,
-                                       tearDown=ZODB.tests.util.tearDown))
+    suite.addTest(doctest.DocTestSuite(
+        setUp=zope.testing.setupstack.setUpDirectory,
+        tearDown=zope.testing.setupstack.tearDown))
+    suite.addTest(ZODB.tests.testblob.storage_reusable_suite(
+        'BlobFileStorage',
+        lambda name, blob_dir:
+        ZODB.FileStorage.FileStorage('%s.fs' % name, blob_dir=blob_dir),
+        test_blob_storage_recovery=True,
+        test_packing=True,
+        ))
+    suite.addTest(PackableStorage.IExternalGC_suite(
+        lambda : ZODB.FileStorage.FileStorage(
+            'data.fs', blob_dir='blobs', pack_gc=False)))
     return suite
 
 if __name__=='__main__':
